from actionsheets.sheets import Actionsheets, default_sheets
from actionsheets.sheet import ActionsheetView
import polars as pl

from rich.console import RenderResult, Group, group
from rich.syntax import Syntax
from rich.text import Text
from rich.markdown import Markdown
from rich.table import Table
from rich.panel import Panel
from rich import print, box

_header_color = 'magenta'
_path_color = 'grey50'
_description_color = 'hot_pink'
_details_color = 'cyan'
_action_color = 'sea_green1'
_bg_color2 = 'grey15'


def print_sheets(sheets: Actionsheets, parent_id: str = ''):
    print(_render_sheets(sheets, parent_id=parent_id))


def print_sheet(sheets: Actionsheets, id: str):
    print(_render_sheet(sheets, id))


def print_snippets(snippets: pl.DataFrame, limit=10):
    print(_render_snippets(snippets.head(n=limit)))


@group()
def _render_sheets(sheets: Actionsheets, parent_id: str) -> RenderResult:
    for sheet_id in sheets.sheets(parent=parent_id, nested=False):
        sheet_render = _render_sheet(sheets, id=sheet_id)
        sheets_render = _render_sheets(sheets, parent_id=sheet_id)
        yield Group(sheet_render, sheets_render)


@group()
def _render_sheet(sheets: Actionsheets, id: str) -> RenderResult:
    info = sheets.sheet_info(sheet=id)
    view = sheets.sheet_view(sheet=id)

    sheet_path = ' > '.join([v.capitalize() for v in info['sheet_id'].split('.')])

    yield Panel(
        Text.assemble(
            (Text(info['title'])),
            (' Actionsheet', 'i'),
            '    ',
            (sheet_path, f'{_path_color}')
        ),
        box=box.DOUBLE_EDGE,
        style=_header_color
    )

    if info['description']:
        yield Text('DESCRIPTION', style=f'{_description_color} bold')
        yield Markdown(markup=info['description'], style=_description_color,
                       inline_code_lexer=info['language'])
        yield ''

    if info['details']:
        yield Text('DETAILS', style=f'{_details_color} bold')
        yield Markdown(markup=info['details'], style=_details_color,
                       inline_code_lexer=info['language'])
        yield ''

    yield _render_sections(view, section='')


@group()
def _render_sections(view: ActionsheetView, section: str) -> RenderResult:
    for section_id in view.entries(parent=section, type='section'):
        yield _render_section(view, section=section_id)


@group()
def _render_section(view: ActionsheetView, section: str) -> RenderResult:
    info = view.section_info(section=section)

    sheet_path = ' > '.join([v.capitalize() for v in info['sheet_id'].split('.')])
    parent_path = ' > '.join([v.capitalize() for v in info['parent_section'].split('.')])
    has_snippets = view.section_snippets(section=section).height > 0

    md = Markdown(
        markup=info['description'],
        style=_description_color,
        inline_code_lexer=info['language']
    )
    snippets_group = _render_section_snippets(view, section)
    yield Panel(
        Group(md, snippets_group),
        title=Text.assemble(
            ('' if info['parent_section'] == '' else f' {parent_path} >',
             f'black on {_header_color} bold'),
            (f' {info["title"]} ', f'white on {_header_color} bold'),
            ' ── ',
            (f'Section {section}', 'black on default')
        ),
        title_align='left',
        subtitle='' if not has_snippets else Text.assemble(
            (f'Actionsheet {sheet_path} ', f'{_path_color} italic'),
            ('───', 'white'),
            (f' Section {section}', f'{_path_color} italic')
        ),
        subtitle_align='left'
    )

    # Render subsections
    yield _render_sections(view, section=section)


def _render_section_snippets(view: ActionsheetView, section: str) -> RenderResult:
    data = view.section_snippets(section=section)
    if data.height:
        return _render_snippets(data)
    else:
        return Group()


def _render_snippets(data: pl.DataFrame) -> RenderResult:
    table = Table(
        collapse_padding=False,
        pad_edge=False,
        show_edge=False,
        show_lines=False,
        expand=True,
        row_styles=['', f'on {_bg_color2}']
    )

    table.add_column('Action', style=_action_color, min_width=15)
    table.add_column('Code', no_wrap=True, overflow='fold', min_width=40, max_width=100)
    table.add_column('Details', style=f'{_details_color} i', min_width=10)

    for snippet in data.iter_rows(named=True):
        table.add_row(
            Markdown(snippet['title']),
            Syntax(
                code=snippet['code'],
                lexer=snippet['language'],
                code_width=120,
                tab_size=2,
                background_color='default' if table.row_count % 2 == 0 else _bg_color2
            ),
            Markdown(snippet['details']),
            style='' if table.row_count % 2 == 0 else f'on {_bg_color2}'
        )

    return table


if __name__ == '__main__':
    print('== Actionsheets.console package ==')
    print_sheets(default_sheets())



import warnings
from importlib import resources
from importlib.resources.abc import Traversable
from typing import Iterator, Self

import polars as pl
import re

from actionsheets.sheet import ActionsheetView, parse_toml_file


class Actionsheets:
    def __init__(self, sheets_data: pl.DataFrame, snippets_data: pl.DataFrame):
        self.sheets_data = sheets_data
        self.snippets_data = snippets_data

    def sheets(self, parent: str = '', nested: bool = True) -> list[str]:
        """
        Get the sheet IDs belonging to the given parent sheet (or root)
        :param parent: Sheet ID
        :param nested: Whether to include nested children in the list
        :return: List of sheet IDs
        """
        if nested:
            return self._all_sheet_ids(parent=parent)
        else:
            return self._child_sheet_ids(parent=parent)

    def __len__(self) -> int:
        """
        Get the total number of sheets
        :return: Number of sheets
        """
        return self.count_sheets()

    def count_sheets(self, parent: str = '', nested: bool = True) -> int:
        """
        Count the number of sheets under the given parent sheet
        :param parent: Parent sheet ID (optional)
        :param nested: Whether to count nested sheets
        :return: Number of sheets
        """
        return len(self.sheets(parent=parent, nested=nested))

    def count_snippets(self, parent: str = '', nested: bool = True) -> int:
        """
        Count the number of snippets under the given parent sheet
        :param parent: Parent sheet ID (optional)
        :param nested: Whether to count snippets from nested sheets
        :return: Number of snippets
        """
        sheet_ids = self.sheets(parent=parent, nested=nested)

        return self.snippets_data.filter(
            (pl.col('sheet').is_in(sheet_ids)) & (pl.col('type') == 'action')
        ).height

    def _all_sheet_ids(self, parent: str = '') -> list[str]:
        return self.sheets_data.filter(
            pl.col('sheet').str.starts_with(parent)
        )['sheet'].to_list()

    def _child_sheet_ids(self, parent: str = '') -> list[str]:
        return self.sheets_data.filter(
            pl.col('sheet_parent') == parent
        )['sheet'].to_list()

    def has_sheet(self, sheet: str, parent: str = '', nested: bool = True) -> bool:
        """
        Check whether the sheet is defined
        :param sheet: Sheet ID
        :param parent: Parent sheet ID (optional)
        :param nested: Whether to search for nested sheets
        :return: Whether the sheet is defined
        """
        return sheet in self.sheets(parent=parent, nested=nested)

    def sheet_info(self, sheet: str) -> dict:
        """
        Get info about a sheet
        :param sheet: Sheet ID
        :return: Dictionary with sheet info
        """
        assert sheet in self.sheets_data['sheet'], \
            f'attempted to access data of undefined sheet "{sheet}"'
        info = self.sheets_data.row(by_predicate=pl.col('sheet') == sheet, named=True)

        info['parents'] = sheet.split(sep='.')[:-1]

        return info

    def sheet_view(self, sheet: str) -> ActionsheetView:
        """
        Create an actionsheet view for the given sheet
        :return: View restricted to this sheet
        """
        return ActionsheetView(
            info=self.sheet_info(sheet),
            data=self.snippets_data.filter(pl.col('sheet') == sheet)
        )

    def filter(self, sheet: str) -> Self:
        sheets = self.sheets(parent=sheet) + [sheet]

        return Actionsheets(
            sheets_data=self.sheets_data.filter(pl.col('sheet').is_in(sheets)),
            snippets_data=self.snippets_data.filter(pl.col('sheet').is_in(sheets))
        )

    def find_sheet(self, query: str) -> str:
        """
        Find the best-matching sheet ID for the given query
        :param query: Search query
        :return: Sheet ID, or empty string if no matches
        """
        terms = re.split(r'\s+|,|\.|\|', query)

        result = self.sheets_data.with_columns(
            pl.col('sheet').str.count_matches('|'.join(terms)).alias('matches')
        ).filter(pl.col('matches') > 0)

        if result.height:
            return result.sort('matches', descending=True).head(n=1)[0, 'sheet']
        else:
            return ''

    def find_snippets(self, query: str, limit: int = 10) -> Self:
        """
        Find snippets across sheets
        :param query: Search query
        :param limit: Max number of snippets in the result
        :return: Actionsheets collection of matches. May be empty in case of no matches.
        """
        terms = re.split(r'\s+|,|\.|\|', query)
        search_pattern = '|'.join(terms)

        result = self.snippets_data.filter(pl.col('type') == 'action').with_columns(
            (
                    pl.col('entry').str.count_matches(search_pattern) +
                    pl.col('sheet').str.count_matches(search_pattern)
            ).alias('matches')
        ).filter(pl.col('matches') > 0)

        filtered_data = (
            result.sort('matches', descending=True).
            head(n=limit).
            select(pl.exclude('matches'))
        )

        return Actionsheets(self.sheets_data, filtered_data)


def _gather_default_files() -> list[str]:
    data_root = resources.files('actionsheets.data')

    def gather_files(entries: Iterator[Traversable]) -> list[str]:
        files_list = []
        for entry in entries:
            if entry.is_dir():
                files_list += gather_files(entry.iterdir())
            elif entry.name.endswith('.toml'):
                files_list += [entry]
        return files_list

    files = gather_files(data_root.iterdir())

    assert files, 'no TOML topic files found inside the data subpackage'
    return files


def parse_toml_files(files: list[str]) -> Actionsheets:
    sheet_info_list = []
    sheet_data_list = []
    for file in files:
        print(f'Parsing file {file}...')
        sheet = parse_toml_file(file)
        print(f'Parsed topic: {sheet.info["name"]}')

        sheet_info_list.append(sheet.info)
        sheet_data_list.append(sheet.data)

    return _process_sheet_list(sheet_info_list, sheet_data_list)


def _process_sheet_list(
        sheet_info_list: list[dict],
        sheet_data_list: list[pl.DataFrame]
) -> Actionsheets:
    sheets_data = _process_sheets(pl.DataFrame(sheet_info_list))
    snippets_data = _process_snippets(pl.concat(sheet_data_list, how='diagonal'))

    return Actionsheets(sheets_data, snippets_data)


def _process_sheets(sheets: pl.DataFrame) -> pl.DataFrame:
    # Rename columns
    sheets = sheets.rename({'name': 'sheet_name', 'parent': 'sheet_parent'})

    # Generate IDs
    sheets = sheets.with_columns(
        pl.when(pl.col('sheet_parent') == '').
        then(pl.col('sheet_name')).
        otherwise(pl.col('sheet_parent') + '.' + pl.col('sheet_name')).
        alias('sheet')
    )

    # Check for missing parent topics
    missing_sheet_names = (
        sheets['sheet_parent'].
        filter(sheets['sheet_parent'].is_in(sheets['sheet']).not_()).
        replace('', None).
        drop_nulls()
    )
    assert missing_sheet_names.is_empty(), \
        f'missing definition for parent topic(s): {", ".join(missing_sheet_names)}'

    # Fill optional fields, compute depth
    sheets = sheets.with_columns(
        pl.col('description').fill_null(''),
        pl.col('details').fill_null(''),
        pl.col('sheet').str.count_matches(r'\.').alias('depth')
    )

    # Set column order
    col_order = ['sheet', 'sheet_parent', 'sheet_name', 'language']
    return sheets.select(pl.col(col_order), pl.exclude(col_order))


def _process_snippets(snippets_data: pl.DataFrame) -> pl.DataFrame:
    # compute derived columns
    snippets_data = snippets_data.with_columns(
        pl.col('title').fill_null(pl.col('entry')),
        pl.col('details').fill_null('').str.strip_chars_end(),
        pl.col('description').fill_null('').str.strip_chars_end(),
        pl.col('entry').str.count_matches('.', literal=True).alias('depth'),
        (
            pl.when(pl.col('sheet_parent') == '').
            then(pl.col('sheet_name')).
            otherwise(pl.col('sheet_parent') + '.' + pl.col('sheet_name'))
        ).alias('sheet')
    )

    # check for code exceeding 80 chars
    long_snippets = (
        snippets_data.
        select(
            pl.col('sheet'),
            pl.col('entry'),
            pl.col('code').str.split(by='\n').alias('code_lines')
        ).
        drop_nulls(subset='code_lines').
        explode(columns='code_lines').
        filter(
            pl.col('code_lines').str.len_chars() > 80
        ).
        select(pl.exclude('code_lines'))
    )

    if long_snippets.height > 0:
        with pl.Config() as cfg:
            cfg.set_tbl_hide_dataframe_shape(True)
            cfg.set_tbl_hide_column_data_types(True)
            cfg.set_fmt_str_lengths(100)
            warnings.warn(
                f'{long_snippets.height} code snippet(s) exceed max width of 80 chars:\n{long_snippets}'
            )

    # ensure column order
    col_order = ['sheet', 'sheet_parent', 'sheet_name', 'entry']
    return snippets_data.select(pl.col(col_order), pl.exclude(col_order))


_default_sheets: Actionsheets = None


def default_sheets() -> Actionsheets:
    global _default_sheets

    if _default_sheets is None:
        files = _gather_default_files()
        _default_sheets = parse_toml_files(files)

    return _default_sheets


if __name__ == '__main__':
    sheets = default_sheets()



import itertools
import re
import tomllib
from functools import reduce
from typing import Literal, Self

import polars as pl

header_keys = ('name', 'language', 'parent', 'title', 'description', 'details', 'code')
section_keys = ('section', 'description', 'details', 'code')
solution_keys = ('code', 'details')
action_keys = tuple(['action', 'description'] + list(solution_keys))
entry_keys = tuple(set(section_keys + action_keys + solution_keys))
reserved_keys = ('name', 'id', 'depth')


class ActionsheetView:
    def __init__(self, info: dict, data: pl.DataFrame):
        self.info = info
        self.data = data

    def __len__(self) -> int:
        """
        Get the number of snippets contained in this view
        :return: Number of snippets
        """
        return self.snippets().height

    def entries(
            self,
            type: Literal['section', 'part', 'action'],
            parent: str = '',
            nested: bool = True
    ) -> list[str]:
        """
        Get the IDs of all entries belonging to the given section
        :param type: Entry type
        :param parent: Parent entry (optional)
        :param nested: Whether to include nested entries
        :return: List of IDs
        """
        if nested:
            return self.data.filter(
                (pl.col('parent_entry').str.starts_with(parent)) & (pl.col('type') == type)
            )['entry'].to_list()
        else:
            return self.data.filter(
                (pl.col('parent_entry') == parent) & (pl.col('type') == type)
            )['entry'].to_list()

    def count_snippets(self, parent: str = '', nested: bool = True) -> int:
        """
        Count the number of snippets under the given parent entry
        :param parent: Parent entry ID (optional)
        :param nested: Whether to count snippets of nested entries
        :return: Number of snippets
        """
        return len(self.entries(type='action', parent=parent, nested=nested))

    def has_section(self, section: str) -> bool:
        """
        Check whether the section is defined in the view
        :param section: Section ID
        :return: Whether the section exists
        """
        return section in self.data.filter(pl.col('type') == 'section')['entry']

    def has_snippet(self, snippet: str) -> bool:
        """
        Check whether the snippet is defined in the view
        :param snippet: Snippet ID
        :return: Whether the snippet exists in the view
        """
        return snippet in self.data.filter(pl.col('type') == 'action')['entry']

    def section_info(self, section: str) -> dict:
        """
        Get metadata about a section
        :param section: Section ID
        :return: Dictionary with info
        """
        info = self.data.row(
            by_predicate=(pl.col('type') == 'section') & (pl.col('entry') == section),
            named=True
        )

        info['parents'] = section.split(sep='.')[:-1]
        return info

    def filter(self, entries: list[str]) -> Self:
        """
        Get a filtered view comprising the given entries, and all of their parents
        :param entries: List of entry IDs to include in the reduced view
        :return: The filtered view
        """
        all_parents = set(itertools.chain.from_iterable(
            [self.get_parents(e) for e in entries]
        ))
        print(all_parents)

        df_parents = self.data.filter(
            pl.col('entry').is_in(all_parents)
        )
        df_entries = self.data.filter(
            pl.col('entry').is_in(entries)
        )
        filtered_data = pl.DataFrame.vstack(
            df_parents,
            df_entries
        ).unique(subset='entry')

        return ActionsheetView(info=self.info, data=filtered_data)

    def filter_section(self, section: str) -> Self:
        """
        Get a filtered view comprising only the children of a given section
        :param section: Section ID
        :return: The filtered view
        """
        return ActionsheetView(
            info=self.info,
            data=self.data.filter(pl.col('entry').str.starts_with(section))
        )

    def section_snippets(self, section: str) -> pl.DataFrame:
        """
        Get the snippets data of a section
        :param section: Section ID
        :return: Snippets data
        """
        return self.snippets().filter(pl.col('parent_entry') == section)

    def snippets(self) -> pl.DataFrame:
        """
        Get all snippets data
        :return: Snippets data
        """
        return self.data.filter(pl.col('type') == 'action')

    def find_snippets(self, query: str, limit: int = 10) -> pl.DataFrame:
        """
        Search for snippets
        :param query: Search query
        :param limit: Max number of snippets in the result
        :return: Snippets data for matching snippets. May be empty in case of no matches.
        """
        terms = re.split(r'\s+|,|\.|\|', query)

        result = self.snippets().with_columns(
            pl.col('entry').str.count_matches('|'.join(terms)).alias('matches')
        ).filter(pl.col('matches') > 0)

        return (
            result.sort('matches', descending=True).
            head(n=limit).
            select(pl.exclude('matches'))
        )

    @staticmethod
    def get_parent(entry: str) -> str:
        """
        Get the parent ID for the given entry
        :param entry: Entry ID
        :return: Parent entry ID, or '' if no parent
        """
        parent_names = entry.split(sep='.')[:-1]
        return '.'.join(parent_names)

    @staticmethod
    def get_parents(entry: str) -> list[str]:
        """
        Get a list of parent IDs for the given entry, up to the root
        :param entry: Entry ID
        :return: List of parent entry IDs
        """
        parent_names = entry.split(sep='.')[:-1]
        return reduce(
            lambda path, p: path + [path[-1] + '.' + p] if path else [p],
            parent_names,
            []
        )


def parse_toml_file(path) -> ActionsheetView:
    """
    Parse an actionsheet TOML file
    :param path: The path to open
    :return: Processed action sheet info and data
    """
    with open(path) as file:
        file_content = file.read()
        return parse_toml(content=file_content, content_id=path)


def parse_toml(content: str, content_id: str = 'undefined') -> ActionsheetView:
    """
    Parse an actionsheet TOML string
    :param content: The TOML content to parse
    :param content_id: Identifier for this content in source reporting
    :return: Processed action sheet info and data
    """
    data = tomllib.loads(content)

    assert data, f'{content_id} has no TOML content'

    # Process header content
    sheet_info = _process_header(data=data, content_id=content_id)

    # Process sections & snippets
    sheet_data = _process_body(data, name=sheet_info['name'], id='').with_columns(
        pl.lit(sheet_info['name']).alias('sheet_name'),
        pl.lit(sheet_info['parent']).alias('sheet_parent'),
        pl.lit(sheet_info['language']).alias('language')
    )

    return ActionsheetView(info=sheet_info, data=sheet_data)


def _process_header(data: dict, content_id: str) -> dict:
    assert 'name' in data, f'{content_id}: no name defined'
    assert_name(data['name'], content_id)

    assert 'parent' in data, f'no parent defined'
    assert isinstance(data['parent'], str), f'parent must be str'

    assert 'language' in data, f'no language defined'
    assert isinstance(data['language'], str), f'language must be str'
    assert data['language'], f'language is empty'

    assert 'title' in data, f'no title defined'
    assert isinstance(data['title'], str), f'title must be str'
    assert data['title'], f'title is empty'

    assert 'description' not in data or isinstance(data['description'], str), \
        f'description must be str'
    assert 'details' not in data or isinstance(data['details'], str), f'details must be str'

    sheet_info = {k: data[k] for k in data.keys() if k in header_keys}
    return sheet_info


def _process_body(content: dict, name: str, id: str) -> pl.DataFrame:
    return _process_entries(content, id=id, parent_entry=dict(), parent_section='')


def _process_entries(
        content: dict,
        id: str,
        parent_entry: dict,
        parent_section: str
) -> pl.DataFrame:
    df = pl.DataFrame()

    entries = _get_entries(content)
    for entry_name in entries:
        entry_id = id + '.' + entry_name if id else entry_name
        entry_dict = _process_entry(
            data=content[entry_name],
            name=entry_name,
            id=entry_id,
            parent_data=parent_entry,
            parent_entry=parent_section
        )

        df2 = _process_entries(
            content=content[entry_name],
            id=entry_id,
            parent_entry=entry_dict,
            parent_section=entry_id if entry_dict['type'] == 'section' else parent_section
        )

        df = pl.concat([df, pl.DataFrame(entry_dict), df2], how='diagonal_relaxed')

    return df


def _get_entries(content: dict) -> list[str]:
    return [k for k in content if isinstance(content[k], dict)]


def _process_entry(
        data: dict,
        name: str,
        id: str,
        parent_data: dict,
        parent_entry: str
) -> dict:
    for k in reserved_keys:
        assert k not in data, f'reserved field "{k}" used in entry {id}'

    entry_dict = {k: data[k] for k in data.keys() if k in entry_keys}

    is_virtual = len(entry_dict) == 0
    is_section = 'section' in entry_dict
    is_action = 'action' in entry_dict
    is_solution = not is_action and 'code' in entry_dict

    assert sum([is_virtual, is_section, is_action, is_solution]) == 1, \
        f'entry {id} is ambiguous; invalid set of fields: {", ".join(entry_dict)}'
    assert is_solution ^ (is_section or is_action or is_virtual)

    entry_dict['entry'] = id
    entry_dict['parent_entry'] = parent_entry

    if is_action:
        return _process_action(entry_dict, name=name, id=id)
    elif is_solution:
        return _process_solution(entry_dict, name=name, id=id, parent_entry=parent_data)
    elif is_section:
        return _process_section(entry_dict, name=name, id=id)
    elif is_virtual:
        return _process_virtual_section(entry_dict, name=name, id=id)


def _process_section(entry: dict, name: str, id: str) -> dict:
    if 'section' in entry:
        assert isinstance(entry['section'], str), f'"section" must be str for entry {id}'
        assert entry['section'], f'no text for "section" defined for entry {id}'
    else:
        entry['section'] = name.capitalize()

    entry['type'] = 'section'
    entry['title'] = entry.pop('section')
    return entry


def _process_virtual_section(entry: dict, name: str, id: str) -> dict:
    entry['type'] = 'part'
    return entry


def _process_action(entry: dict, name: str, id: str) -> dict:
    assert 'action' in entry, f'no "action" defined for entry {id}'
    assert isinstance(entry['action'], str), f'"action" must be str for entry {id}'
    assert entry['action'], f'no text for "action" defined for entry {id}'

    assert 'code' in entry, f'no "code" defined for entry {id}'
    assert isinstance(entry['code'], str), f'"code" must be str for entry {id}'
    assert entry['code'], f'no text for "code" defined for entry {id}'

    entry['type'] = 'action'
    entry['title'] = entry.pop('action')
    return entry


def _process_solution(entry: dict, name: str, id: str, parent_entry: dict) -> dict:
    assert parent_entry, f'empty parent action entry for solution {id}'
    assert 'title' in parent_entry, f'missing title for parent of solution {id}'
    entry['title'] = parent_entry['title']
    entry['type'] = 'action'
    return entry


def assert_name(name: any, path: str):
    assert isinstance(name, str), f'{path}: name must be str'
    assert name, f'{path}: name is empty'
    assert '.' not in name, f'{path}: name cannot contain "."'
